@@ -30,7 +30,11 @@
 #include "xf86.h"
 
 /* for usleep */
-#include "xf86_ansic.h"
+#if HAVE_XF86_ANSIC_H
+# include "xf86_ansic.h"
+#else
+# include <unistd.h>
+#endif
 
 #include "rhd.h"
 #include "rhd_crtc.h"
@@ -48,6 +52,7 @@ struct rhdCrtcStore {
     CARD32 GrphYStart;
     CARD32 GrphXEnd;
     CARD32 GrphYEnd;
+    CARD32 GrphSwap;
     CARD32 GrphPrimarySurfaceAddress;
     CARD32 GrphSurfaceOffsetX;
     CARD32 GrphSurfaceOffsetY;
@@ -164,7 +169,7 @@ DxFBSet(struct rhdCrtc *Crtc, CARD16 Pitch, CARD16 Width, CARD16 Height,
     RHDPtr rhdPtr = RHDPTRI(Crtc);
     CARD16 RegOff;
 
-    RHDDebug(Crtc->scrnIndex, "%s: %s\n", __func__, Crtc->Name);
+    RHDDebug(Crtc->scrnIndex, "FUNCTION: %s: %s\n", __func__, Crtc->Name);
 
     if (Crtc->Id == RHD_CRTC_1)
 	RegOff = D1_REG_OFFSET;
@@ -173,24 +178,32 @@ DxFBSet(struct rhdCrtc *Crtc, CARD16 Pitch, CARD16 Width, CARD16 Height,
 
     RHDRegMask(Crtc, RegOff + D1GRPH_ENABLE, 1, 0x00000001);
 
+    /* disable R/B swap, disable tiling, disable 16bit alpha, etc. */
+    RHDRegWrite(Crtc, RegOff + D1GRPH_CONTROL, 0);
+
     switch (bpp) {
     case 8:
-	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0, 0xF10703);
+	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0, 0x00000703);
 	break;
     case 15:
-	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0x000001, 0xF10703);
+	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0x000001, 0x00000703);
 	break;
     case 16:
-	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0x000101, 0xF10703);
+	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0x000101, 0x00000703);
 	break;
     case 24:
     case 32:
     default:
-	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0x000002, 0xF10703);
+	RHDRegMask(Crtc, RegOff + D1GRPH_CONTROL, 0x000002, 0x00000703);
 	break;
     /* TODO: 64bpp ;p */
     }
 
+    /* Make sure that we are not swapping colours around */
+    if (rhdPtr->ChipSet > RHD_R600)
+	RHDRegWrite(Crtc, RegOff + D1GRPH_SWAP_CNTL, 0);
+    /* R5xx - RS690 case is GRPH_CONTROL bit 16 */
+
     RHDRegWrite(Crtc, RegOff + D1GRPH_PRIMARY_SURFACE_ADDRESS,
 		rhdPtr->FbIntAddress + Offset);
     RHDRegWrite(Crtc, RegOff + D1GRPH_PITCH, Pitch);
@@ -202,7 +215,7 @@ DxFBSet(struct rhdCrtc *Crtc, CARD16 Pitch, CARD16 Width, CARD16 Height,
     RHDRegWrite(Crtc, RegOff + D1GRPH_Y_END, Height);
 
     /* D1Mode registers */
-    RHDRegWrite(rhdPtr, RegOff + D1MODE_DESKTOP_HEIGHT, Height);
+    RHDRegWrite(Crtc, RegOff + D1MODE_DESKTOP_HEIGHT, Height);
 
     Crtc->Pitch = Pitch;
     Crtc->Width = Width;
@@ -279,7 +292,7 @@ DxModeSet(struct rhdCrtc *Crtc, DisplayModePtr Mode)
     CARD16 BlankStart, BlankEnd;
     CARD16 RegOff;
 
-    RHDDebug(Crtc->scrnIndex, "%s: %s\n", __func__, Crtc->Name);
+    RHDDebug(Crtc->scrnIndex, "FUNCTION: %s: %s\n", __func__, Crtc->Name);
 
     if (Crtc->Id == RHD_CRTC_1)
 	RegOff = D1_REG_OFFSET;
@@ -583,6 +596,8 @@ DxSave(struct rhdCrtc *Crtc)
     Store->GrphYStart = RHDRegRead(Crtc, RegOff + D1GRPH_Y_START);
     Store->GrphXEnd = RHDRegRead(Crtc, RegOff + D1GRPH_X_END);
     Store->GrphYEnd = RHDRegRead(Crtc, RegOff + D1GRPH_Y_END);
+    if (RHDPTRI(Crtc)->ChipSet >= RHD_R600)
+	Store->GrphSwap = RHDRegRead(Crtc, RegOff + D1GRPH_SWAP_CNTL);
     Store->GrphPrimarySurfaceAddress =
 	RHDRegRead(Crtc, RegOff + D1GRPH_PRIMARY_SURFACE_ADDRESS);
     Store->GrphSurfaceOffsetX =
@@ -657,6 +672,8 @@ DxRestore(struct rhdCrtc *Crtc)
     RHDRegWrite(Crtc, RegOff + D1GRPH_Y_START, Store->GrphYStart);
     RHDRegWrite(Crtc, RegOff + D1GRPH_X_END, Store->GrphXEnd);
     RHDRegWrite(Crtc, RegOff + D1GRPH_Y_END, Store->GrphYEnd);
+    if (RHDPTRI(Crtc)->ChipSet >= RHD_R600)
+	RHDRegWrite(Crtc, RegOff + D1GRPH_SWAP_CNTL, Store->GrphSwap);
     RHDRegWrite(Crtc, RegOff + D1GRPH_PRIMARY_SURFACE_ADDRESS,
 		Store->GrphPrimarySurfaceAddress);
     RHDRegWrite(Crtc, RegOff + D1GRPH_SURFACE_OFFSET_X,
@@ -701,6 +718,15 @@ DxRestore(struct rhdCrtc *Crtc)
     else
 	RHDRegWrite(Crtc, PCLK_CRTC2_CNTL, Store->CrtcPCLKControl);
 
+    /* When VGA is enabled, it imposes its timing on us, so our CRTC SYNC
+     * timing can be set to 0. This doesn't always restore properly...
+     * Workaround is to set a valid sync length for a bit so VGA can
+     * latch in. */
+    if (!Store->CrtcVSyncA && (Store->CrtcControl & 0x00000001)) {
+	RHDRegWrite(Crtc, RegOff + D1CRTC_V_SYNC_A, 0x00040000);
+	usleep(300000); /* seems a reliable timeout here */
+	RHDRegWrite(Crtc, RegOff + D1CRTC_V_SYNC_A, Store->CrtcVSyncA);
+    }
 }
 
 /*
